/* sqExternalBitmap.c -- access to the C heap.  Author: Ian.Piumarta@inria.fr
 *
 * Last edited: Wed Oct 22 17:38:19 1997 by piumarta (Ian Piumarta) on olrik
 */

#include "sqExternalBitmap.h"

static int *chunks[MAXARRAYS];
static char marked[MAXARRAYS];


/*
 * convenience functions
 */

static int registerChunk(int *chunk)
{
  int i;
  for (i= 0; i < MAXARRAYS; ++i)
    if (!chunks[i]) {
      chunks[i]= chunk;
      return i;
    }
  return -1;
}


static int getChunkSize(/*  IN */ oop extBitmap,
			/* OUT */ int **chunk, int *size)
{
  oop handleObj;
  int handle;

  if (instVarSize(extBitmap) != 1) return false;
  handleObj= instVarAt(extBitmap, 0);
  if (!isIntegerObject(handleObj)) return false;
  handle= integerValueOf(handleObj);
  if (handle < 0 || handle >= MAXARRAYS) return false;

  *chunk= chunks[handle];
  if (!*chunk) return false;

  *size= chunks[handle][0];
  return true;
}


static int getChunkSizeIndex(/*  IN */ oop extBitmap, oop indexObj,
			     /* OUT */ int **chunk, int *size, int *index)
{
  if (!isIntegerObject(indexObj)) return false;
  *index= integerValueOf(indexObj);

  return getChunkSize(extBitmap, chunk, size);
}


/*
 * ExternalBitmap primitives
 */

/* allocate storage for a new ExternalBitmap */
oop externalBitmap_initialize(oop extBitmap, oop byteSizeObj)
{
  int byteSize;
  int *chunk;
  int handle;

  if (!isIntegerObject(byteSizeObj)) return primitiveFail();
  if (instVarSize(extBitmap) != 1) return primitiveFail();
  byteSize= integerValueOf(byteSizeObj);
  chunk= (int *)malloc(sizeof(int) + byteSize);
  if (!chunk) return primitiveFail();
  if ((handle= registerChunk(chunk)) < 0) {
    free(chunk);
    return primitiveFail();
  }
  instVarAtPut(extBitmap, 0, integerObjectOf(handle));
  chunk[0]= byteSize;
  return extBitmap;
}


/* anExternalBitmap byteAt: anIntegerIndex => anInteger */
oop externalBitmap_byteAt(oop extBitmap, oop indexObj)
{
  int index;
  int size;
  int *chunk;

  if (!getChunkSizeIndex(extBitmap, indexObj, &chunk, &size, &index))
    return primitiveFail();

  if (index < 1 || index > size)
    return primitiveFail();

  return integerObjectOf(((char *)&chunk[1])[index - 1]);
}


/* anExternalBitmap byteAt: anIntegerIndex put: anInteger => anInteger */
oop externalBitmap_byteAtPut(oop extBitmap, oop indexObj, oop valueObj)
{
  int index;
  int size;
  int *chunk;
  int value;

  if (!isIntegerObject(valueObj))
    return primitiveFail();
  value= integerValueOf(valueObj);
  if (value < 0 || value > 255)
    return primitiveFail();

  if (!getChunkSizeIndex(extBitmap, indexObj, &chunk, &size, &index))
    return primitiveFail();

  if (index < 1 || index > size)
    return primitiveFail();

  ((char *)&chunk[1])[index - 1]= value;

  return valueObj;
}


/* anExternalBitmap wordAt: anIntegerIndex => anInteger */
oop externalBitmap_wordAt(oop extBitmap, oop indexObj)
{
  int index;
  int size;
  int *chunk;
  int value;

  if (!getChunkSizeIndex(extBitmap, indexObj, &chunk, &size, &index))
    return primitiveFail();

  if (index < 1 || index > size / 4)
    return primitiveFail();

  value= chunk[index];

  if (isIntegerValue(value))
    return integerObjectOf(chunk[index]);

  return primitiveFail();	/* need LargeInteger -- let fail code recover */
}


/* anExternalBitmap wordAt: anIntegerIndex put: anInteger => anInteger */
oop externalBitmap_wordAtPut(oop extBitmap, oop indexObj, oop valueObj)
{
  int index;
  int size;
  int *chunk;
  int value;

  if (!isIntegerObject(valueObj))
    return primitiveFail();
  value= integerValueOf(valueObj);

  if (!getChunkSizeIndex(extBitmap, indexObj, &chunk, &size, &index))
    return primitiveFail();

  if (index < 1 || index > size / 4)
    return primitiveFail();

  chunk[index]= value;

  return valueObj;
}


/* anExternalBitmap basicSize => anIntegerByteSize */
oop externalBitmap_basicSize(oop extBitmap)
{
  oop handleObj;
  int handle;

  if (instVarSize(extBitmap) != 1) return primitiveFail();
  handleObj= instVarAt(extBitmap, 0);
  if (!isIntegerObject(handleObj)) return primitiveFail();
  handle= integerValueOf(handleObj);
  if (handle < 0 || handle >= MAXARRAYS) return primitiveFail();

  if (!chunks[handle]) return primitiveFail();

  return integerObjectOf(chunks[handle][0]);
}


/* ExternalBitmap declareLive: anBitmapOfHandles => anIntegerReclaimedCount */
oop externalBitmap_class_decareLive(oop rcvrClass, oop handleArray)
{
  int i, count, reclaimed;

  if (fetchClassOf(handleArray) != Array)
    return primitiveFail();

  count= indexableSize(handleArray);

  for (i= 0; i < MAXARRAYS; ++i)
    marked[i]= false;

  for (i= 0; i < count; ++i) {
    oop handleObj= objectAt(handleArray, i);
    int handle;
    if (!isIntegerObject(handleObj)) return primitiveFail();
    handle= integerValueOf(handleObj);
    if (handle < 0 || handle >= MAXARRAYS) return primitiveFail();
    marked[handle]= true;
  }

  reclaimed= 0;
  for (i= 0; i < MAXARRAYS; ++i)
    if (chunks[i] && !marked[i]) {
      free(chunks[i]);
      chunks[i]= 0;
      ++reclaimed;
    }

  return integerObjectOf(reclaimed);
}


/* anExternalBitmap copyToBitmap: aBitmap => anExternalBitmap */
oop externalBitmap_copyToBitmap(oop extBitmap, oop bitmap)
{
  int bitmapSize;
  int *chunk, chunkSize;

  if (!isWords(bitmap)) return primitiveFail();
  bitmapSize= indexableSize(bitmap);

  if (!getChunkSize(extBitmap, &chunk, &chunkSize)) return primitiveFail();
  chunkSize/= 4;

  if (chunkSize != bitmapSize) return primitiveFail();

  memcpy(firstIndexableField(bitmap), (char *)(chunk + 1), bitmapSize * 4);

  return extBitmap;
}


/* anExternalBitmap copyFromBitmap: aBitmap => anExternalBitmap */
oop externalBitmap_copyFromBitmap(oop extBitmap, oop bitmap)
{
  int bitmapSize;
  int *chunk, chunkSize;

  if (!isWords(bitmap)) return primitiveFail();
  bitmapSize= indexableSize(bitmap);

  if (!getChunkSize(extBitmap, &chunk, &chunkSize)) return primitiveFail();
  chunkSize/= 4;

  if (chunkSize != bitmapSize) return primitiveFail();

  memcpy((char *)(chunk + 1), firstIndexableField(bitmap), bitmapSize * 4);

  return extBitmap;
}


/* anExternalBitmap bitInvert => anExternalBitmap */
oop externalBitmap_bitInvert(oop extBitmap)
{
  int *chunk, chunkSize;

  if (!getChunkSize(extBitmap, &chunk, &chunkSize)) return primitiveFail();
  chunkSize/= 4;

  while (chunkSize--) (*++chunk) ^= -1;

  return extBitmap;
}
